#include <llvm_common.h>
#include <iostream>
#include <NervJIT.h>
// #include <NervApp.h>

#undef DEBUG_MSG
#undef ERROR_MSG
#undef THROW_MSG

#define DEBUG_MSG(msg) std::cout << "[DEBUG]: "<<msg<< std::endl;
#define ERROR_MSG(msg) std::cout << "[ERROR]: "<<msg<< std::endl;

#define THROW_MSG(msg)                                                                          \
    {                                                                                           \
        ERROR_MSG(msg); \
        throw std::runtime_error("An exception just occured.");                                                     \
    }

#define CHECK_RET(cond, ret, msg) \
    if (!(cond))                  \
    {                             \
        THROW_MSG(msg);           \
        return ret;               \
    }

// cf. https://stackoverflow.com/questions/54403377/problems-enabling-rtti-in-llvm-jit-ed-code
// #pragma comment(linker, "/export:??_7type_info@@6B@")
// #pragma comment(linker, "/export:??3@YAXPEAX_K@Z")
// #pragma comment(linker, "/export:?_Facet_Register@std@@YAXPEAV_Facet_base@1@@Z")

// __declspec(dllexport) 
int my_function(int a, int b)
{
	return a*b;
}


int main(int argc, char *argv[])
{

	// auto& app = nv::NervApp::instance();

#if 0
	DEBUG_MSG("Running clang compilation...");
    runClang({"W:/Projects/NervSeed/temp/test1.cxx",
			  "W:/Projects/NervSeed/temp/test2.cxx"});
	DEBUG_MSG("Done running clang compilation.");
#else
	DEBUG_MSG("Initializing LLVM...");
	nv::initLLVM(argc, argv);

	DEBUG_MSG("Creating NervJIT...");
	auto jit = std::make_unique<nv::NervJIT>();

	// int res = my_function((int)getenv("v1"),(int)getenv("v2"));
	// DEBUG_MSG("The dummy value is: "<<res);
	// DEBUG_MSG("_MSC_VER="<<_MSC_VER);

	jit->clearHeaderSearchPaths();
	jit->addHeaderSearchPath("W:/Projects/NervSeed/deps/msvc64/llvm-20200409/lib/clang/11.0.0/include", nv::NervJIT::HEADER_SYSTEM);
	jit->addHeaderSearchPath("D:/Apps/VisualStudio2017_CE/VC/Tools/MSVC/14.16.27023/include", nv::NervJIT::HEADER_SYSTEM);
	jit->addHeaderSearchPath("D:/Apps/VisualStudio2017_CE/VC/Tools/MSVC/14.16.27023/atlmfc/include", nv::NervJIT::HEADER_SYSTEM);
	jit->addHeaderSearchPath("C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/ucrt", nv::NervJIT::HEADER_SYSTEM);
	jit->addHeaderSearchPath("C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/shared", nv::NervJIT::HEADER_SYSTEM);
	jit->addHeaderSearchPath("C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/um", nv::NervJIT::HEADER_SYSTEM);
	jit->addHeaderSearchPath("C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/winrt", nv::NervJIT::HEADER_SYSTEM);

#if 0
	// This will not really be working anymore since we are not using our "custom demangling mapping" anymore
	DEBUG_MSG("loading modules");
	jit->loadModuleFromFile("W:/Projects/NervSeed/temp/test1.cxx");
	jit->loadModuleFromFile("W:/Projects/NervSeed/temp/test2.cxx");
	// jit->loadModuleFromFile("W:/Projects/NervSeed/temp/test3.cxx");
	jit->loadModuleFromBuffer(R"(
int nv_add3(int a, int b)
{
    return (a+b)*3;
}

int nv_sub3(int a, int b)
{
    return (a-b)*3;
}
)");

	typedef int(*Func)(int, int);

	Func add1 = (Func)jit->lookup("int nv_add(int, int)");
	CHECK_RET(add1!=nullptr,1,"Invalid nv_add function.");
	DEBUG_MSG("nv_add(12,3) = "<< add1(12,3));
	// Func add2 = (Func)jit->lookup("nv_add2");
	Func add2 = (Func)jit->lookup("int nv_add2(int, int)");
	CHECK_RET(add2!=nullptr,1,"Invalid nv_add2 function.");
	DEBUG_MSG("nv_add(12,3) = "<< add2(12,3));
	Func add3 = (Func)jit->lookup("int nv_add3(int, int)");
	CHECK_RET(add3!=nullptr,1,"Invalid nv_add3 function.");
	DEBUG_MSG("nv_add(12,3) = "<< add3(12,3));
#endif

#if 0
	jit->loadModuleFromFile("W:/Projects/NervSeed/temp/test5.cxx");
	auto func = (int(*)())jit->lookup("test_function");
	CHECK_RET(func!=nullptr,1,"Invalid test_function pointer.");
	DEBUG_MSG("test_function() result: "<<func());
#endif

#if 0
	// jit->addCurrentProcess();
	// jit->addMacroDefinition("_MT");
	jit->addMacroDefinition("_DLL");
	jit->addMacroDefinition("_CRT_SECURE_NO_WARNINGS");
	// jit->addMacroDefinition("CRTDLL");
	// jit->addMacroDefinition("_HAS_EXCEPTIONS");
	// jit->addMacroDefinition("NDEBUG");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/msvcp140.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/msvcp140_2.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/concrt140.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/vcruntime140.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/ucrtbase.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/llvm_helper.dll");
	// jit->addDynamicLib("C:/WINDOWS/system32/msvcrt.dll");
	// jit->addDynamicLib("C:/WINDOWS/system32/MSVCP140.dll");
	// jit->addMacroDefinition("UNICODE");
	// jit->addMacroDefinition("_UNICODE");
	// jit->addMacroDefinition("_WIN32");
	// jit->addMacroDefinition("_MSC_VER=1916");
	// jit->addDynamicLib("C:/WINDOWS/system32/kernel32.dll");
	// jit->addDynamicLib("C:/WINDOWS/system32/VCRUNTIME140.dll");
	// jit->addDynamicLib("api-ms-win-crt-runtime-l1-1-0.dll");
	// jit->addDynamicLib("api-ms-win-crt-string-l1-1-0.dll");
	// jit->addDynamicLib("api-ms-win-crt-heap-l1-1-0.dll");
	// jit->addDynamicLib("api-ms-win-crt-math-l1-1-0.dll");
	// jit->addDynamicLib("api-ms-win-crt-stdio-l1-1-0.dll");
	// jit->addDynamicLib("api-ms-win-crt-locale-l1-1-0.dll");
	
	// jit->addDynamicLib("C:/WINDOWS/system32/CONCRT140.dll");
	// jit->addDynamicLib("C:/WINDOWS/system32/ucrtbase.dll");
	// jit->addDynamicLib("C:/WINDOWS/system32/msvcp140_1.dll");
	// jit->addDynamicLib("C:/WINDOWS/system32/msvcp140_2.dll");
	jit->loadModuleFromFile("W:/Projects/NervSeed/temp/test7.cxx");
	auto func = (int(*)())jit->lookup("test");
	CHECK_RET(func!=nullptr,1,"Invalid test pointer.");
	DEBUG_MSG("test() result: "<<func());
#endif

#if 1
	// jit->addMacroDefinition("_MT");
	jit->addMacroDefinition("_DLL");
	jit->addMacroDefinition("_CRT_SECURE_NO_WARNINGS");

	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/nvCore.dll");

	// jit->addDynamicLib("W:/Windows/system32/msvcp140.dll");
	// jit->addDynamicLib("W:/Windows/system32/msvcp140_2.dll");
	// jit->addDynamicLib("W:/Windows/system32/concrt140.dll");
	// jit->addDynamicLib("W:/Windows/system32/vcruntime140.dll");
	// jit->addDynamicLib("W:/Windows/system32/ucrtbase.dll");

	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/msvcp140.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/msvcp140_2.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/concrt140.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/vcruntime140.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/CRT/ucrtbase.dll");
	jit->addDynamicLib("W:/Projects/NervSeed/dist/bin/msvc64/llvm_helper.dll");

	// Now we load our nvCore related function:
	// jit->addMacroDefinition("_WIN32");
	// jit->addMacroDefinition("_MSC_VER=1916");
	jit->addHeaderSearchPath("W:/Projects/NervSeed/deps/msvc64/boost_1_68_0/include", nv::NervJIT::HEADER_ANGLED);
	jit->addHeaderSearchPath("W:/Projects/NervSeed/sources/nvCore/include", nv::NervJIT::HEADER_ANGLED);
	jit->loadModuleFromFile("W:/Projects/NervSeed/temp/test6.cxx");

	DEBUG_MSG("Retrieving root path function.");
	typedef int(*Func2)();
	Func2 show = (Func2)jit->lookup("showRootPath");
	CHECK_RET(show, 1, "cannot retrieve showRootPath function.");
	DEBUG_MSG("Displaying root path here:");
	int val = show();
	DEBUG_MSG("Result value is: "<<val);

#endif

	DEBUG_MSG("Destroying NervJIT...");
	jit.reset();

	DEBUG_MSG("Uninitializing LLVM...");
	nv::uninitLLVM();
#endif

	// nv::NervApp::destroy();
	// nv::MemoryManager::destroy();

	DEBUG_MSG("Exiting...");
	return 0;
}
